r"""The ID:s of components used for identification and session state management."""

# =====================================================================================
# Create user form
# =====================================================================================

CREATE_USER_FORM = 'create-user-form'
CREATE_USER_FORM_USERNAME_TEXT_INPUT = 'create-user-form-username-text-input'
CREATE_USER_FORM_DISPLAYNAME_TEXT_INPUT = 'create-user-form-displayname-text-input'
CREATE_USER_FORM_AD_USERNAME_TEXT_INPUT = 'create-user-form-ad-username-text-input'
CREATE_USER_FORM_ROLE_SELECTBOX = 'create-user-form-role-selectbox'
CREATE_USER_FORM_CUSTOM_ROLES_MULTISELECTBOX = 'create-user-form-custom-roles-multiselectbox'
CREATE_USER_FORM_EMAIL_TEXT_INPUT = 'create-user-form-email-text-input'
CREATE_USER_FORM_DISABLED_TOGGLE = 'create-user-form-disabled-toggle'
CREATE_USER_FORM_SUBMIT_BUTTON = 'create-user-form-submit-button'

# =====================================================================================
# Delete user
# =====================================================================================

DELETE_USER_BUTTON = 'delete-user-button'

# =====================================================================================
# Register form
# =====================================================================================

BP_REGISTER_FORM = 'bp-register-form'
BP_REGISTER_FORM_USERNAME_TEXT_INPUT = 'bp-register-form-username-text-input'
BP_REGISTER_FORM_DISPLAYNAME_TEXT_INPUT = 'bp-register-form-displayname-text-input'
BP_REGISTER_FORM_EMAIL_TEXT_INPUT = 'bp-register-form-email-text-input'
BP_REGISTER_FORM_CREDENTIAL_NICKNAME_TEXT_INPUT = 'bp-register-form-credential-nickname-text-input'
BP_REGISTER_FORM_DISCOVERABILITY_RADIO_BUTTON = 'bp-register-form-discoverability-radio-button'
BP_REGISTER_FORM_DISCOVERABILITY_TOGGLE_SWITCH = 'bp-register-form-discoverability-toggle-switch'
BP_REGISTER_FORM_ALIASES_TEXT_INPUT = 'bp-register-form-aliases-text-input'
BP_REGISTER_FORM_SUBMIT_BUTTON = 'bp-register-form-submit-button'
BP_REGISTER_FORM_EXISTING_USER_CREDENTIAL_NICKNAME_TEXT_INPUT = (
    'bp-register-form-existing-user-credential-nickname-text-input'
)
BP_REGISTER_FORM_EXISTING_USER_DISCOVERABILITY_RADIO_BUTTON = (
    'bp-register-form-existing-user-discoverability-radio-button'
)
BP_REGISTER_FORM_EXISTING_USER_DISCOVERABILITY_TOGGLE_SWITCH = (
    'bp-register-form-existing-user-discoverability-toggle-switch'
)
BP_REGISTER_FORM_EXISTING_USER_SUBMIT_BUTTON = 'bp-register-form-existing-user-submit-button'

# =====================================================================================
# Sign in form
# =====================================================================================

BP_SIGN_IN_BUTTON = 'bp-sign-in-button'
BP_SIGN_IN_FORM_ALIAS_TEXT_INPUT = 'bp-sign-in-form-alias-text-input'
BP_SIGN_IN_FORM_SUBMIT_BUTTON = 'bp-sign-in-form-submit-button'

# =====================================================================================
# Sign out
# =====================================================================================

KEY_SIGN_OUT_BUTTON = 'stp-sign-out-button'

# =====================================================================================
# Update user form
# =====================================================================================

UPDATE_USER_FORM = 'update-user-form'
UPDATE_USER_FORM_USERNAME_TEXT_INPUT = 'update-user-form-username-text-input'
UPDATE_USER_FORM_DISPLAYNAME_TEXT_INPUT = 'update-user-form-displayname-text-input'
UPDATE_USER_FORM_AD_USERNAME_TEXT_INPUT = 'update-user-form-ad-username-text-input'
UPDATE_USER_FORM_ROLE_SELECTBOX = 'update-user-form-role-selectbox'
UPDATE_USER_FORM_CUSTOM_ROLES_MULTISELECTBOX = 'update-user-form-custom-roles-multiselectbox'
UPDATE_USER_FORM_DISABLED_TOGGLE = 'update-user-form-disabled-toggle'
UPDATE_USER_FORM_SUBMIT_BUTTON = 'update-user-form-submit-button'
