r"""The ID:s (keys) of components used for identification and session state management."""

# =====================================================================================
# Admin
# =====================================================================================

ADMIN_USER_ROLE_MULTISELECT = 'stp-admin-user-role-multiselect'
ADMIN_USER_STATE_RADIO_BUTTON = 'stp-admin-user-state-radio-button'
ADMIN_USER_SELECTBOX = 'stp-admin-user-selectbox'
ADMIN_CREATE_USER_BUTTON = 'stp-admin-create-user-button'
ADMIN_REFRESH_USERS_BUTTON = 'stp-admin-refresh-users-button'
