r"""Metadata about the streamlit-passwordless package."""

# Standard library
from datetime import date

__versiontuple__ = (0, 16, 0)
r"""The version of streamlit-passwordless in a comparable form.
Adheres to `Semantic Versioning <https://semver.org/spec/v2.0.0.html>`_
(MAJOR.MINOR.PATCH).
"""

__version__ = '0.16.0'
r"""The streamlit-passwordless version string."""

__releasedate__ = date(2025, 6, 27)
r"""The release date of the version specified in :data:`__versiontuple__`."""
